/*
   bfsbrowser.c
   Copyright (C) 2001-2002 Tomohide Naniwa. All rights reserved.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Last modified: Feb. 4, 2002.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include "btronfs.h"
#include "tad.h"

#define CHR_JAPAN 0x0021

#define PROMPT "bfsbrowser> "

/* ѿ */
SYSTEM_HEADER system_header;
FILE_HEADER file_header;
int file_id = 0;
int file_start_block = 0;
int fd;
time_t BTRON_TIME;

/* ؿ */
void print_system_header(SYSTEM_HEADER *system_header);
void print_file_name(FILE_HEADER *header);
void read_file_header(int file_id,
		      FILE_HEADER *file_header, int *file_start_block);
void do_info();
void do_quit();
void do_help();
void do_finfo();
void do_record();
void do_cf();
void do_dump();
void do_view();

struct {
  char command[20];
  void (*func)();
} command_tbl[] = {
  {"dinfo", do_info},
  {"finfo", do_finfo},
  {"record", do_record},
  {"cf", do_cf},
  {"taddump", do_dump},
  {"tadview", do_view},
  {"help", do_help},
  {"quit", do_quit},
  {{0}, NULL}
};

main(int argc, char *argv[])
{
  int i;
  char command[40];
  struct tm btm;

  if (argc < 2) {
    printf("usage: %s 'file name'\n", argv[0]);
    exit(0);
  }

  if ((fd = open(argv[1], O_RDWR)) == -1) {
    perror(argv[1]);
    exit(0);
  }

  /* ƥإåɤ߹ */
  lseek(fd, BLOCK_SIZE, SEEK_SET);
  read(fd, &system_header, sizeof(SYSTEM_HEADER));

  if ((system_header.disk_id != 0x42FE) && (system_header.disk_id != 0x52FE)) {
    printf("File '%s' is not BTRON floppy disk\n", argv[1]);
    exit(1);
  }

  btm.tm_sec = 0;
  btm.tm_min = 0;
  btm.tm_hour = 0;
  btm.tm_mday = 1;
  btm.tm_mon = 0;
  btm.tm_year = 85;
  BTRON_TIME = mktime(&btm);
#ifdef notdef
  printf("BTRON_TIME = %d\n", BTRON_TIME);
  BTRON_TIME = 473353200;
#endif

  /* root file ɤ߹ */
  file_start_block
    = 1024 +
    (system_header.block_size + system_header.id_tbl_size
     + system_header.short_name_tbl_size) *
    system_header.logical_block_size;
  lseek(fd, file_start_block, SEEK_SET);
  read(fd, &file_header, sizeof(FILE_HEADER));

  while(1) {
    printf(PROMPT);
    fgets(command, 40, stdin);
    command[strlen(command)-1] = 0;
    for(i = 0; command_tbl[i].command[0] != 0; ++i) {
      if (strcmp(command, command_tbl[i].command) == 0) {
	(command_tbl[i].func)();
	break;
      }
    }
  }
}

void printk(unsigned short int wchar)
{
  wchar |= 0x8080;
  printf("%c%c", wchar >> 8, wchar & 0x0ff);
}

void print_system_header(SYSTEM_HEADER *header)
{
  int i;
  time_t clock;

  printf("disk id = 0x%x\n", header->disk_id);
  printf("disk type = 0x%x\n", header->disk_type);
  printf("block size = %d\n", header->block_size);
  printf("id tbl size = %d\n", header->id_tbl_size);
  printf("short name tbl size = %d\n", header->short_name_tbl_size);
  printf("bitmap size = %d\n", header->bitmap_size);
  printf("logical block size = %d\n", header->logical_block_size);
  printf("max file = %d\n", header->max_file);
  printf("lang = 0x%x\n", header->lang);
  printf("access level = %d\n", header->access_level);
  printf("logical block = %d\n", header->logical_block);
  printf("empty bolck = %d\n", header->empty_block);
  clock = header->modified_date + BTRON_TIME;
  printf("modified %s", asctime(gmtime(&clock)));
  clock = header->created_date + BTRON_TIME;
  printf("created %s", asctime(gmtime(&clock)));
  if (header->lang == CHR_JAPAN) {
    for(i = 0; i < 20; ++i) {
      if (header->file_system.wchar[i] != 0) {
	printk(header->file_system.wchar[i]);
      }
      else {
	printf("\n");
	break;
      }
    }
    for(i = 0; i < 20; ++i) {
      if (header->device_name.wchar[i] != 0) {
	printk(header->device_name.wchar[i]);
      }
      else {
	printf("\n");
	break;
      }
    }
  }
}

void print_file_name(FILE_HEADER *header)
{
  int i;

  if (system_header.lang == CHR_JAPAN) {
    for(i = 0; i < 20; ++i) {
      if (header->file_name.wchar[i] != 0) {
	printk(header->file_name.wchar[i]);
      }
      else {
	break;
      }
    }
  }
}

void print_file_header(FILE_HEADER *header)
{
  printf("file header id = 0x%x\n", header->header_start_id);
  printf("file type = 0x%04x\n", header->file_type);
  printf("application type = 0x%x\n", header->application_type);
  printf("link number = %d\n", header->link_number);
  printf("index level = %d\n", header->index_level);
  printf("byte = %d\n", header->byte);
  printf("logical block = %d\n", header->logical_block);
  printf("record = %d\n", header->record);
  print_file_name(header); printf("\n");
  printf("file id = %d\n", header->file_id);
  printf("file end id = 0x%x\n", header->file_end_id);
}

void print_record_index(RECORD_INDEX *rec)
{
  int i;
  FILE_HEADER fh;

  if (rec->count == 0) {
    if (rec->normal.record_type == 0) {
      printf("\n");
      return;
    }
    else if (rec->normal.record_type == RT_LINK) {
      printf("record type = 0x%02x\t", rec->link.record_type);
#if 0
      printf("record sub type = 0x%04x\t", rec->link.record_sub_type);
#endif
      printf("file id = %d\t", rec->link.file_id);
      read_file_header(rec->link.file_id, &fh, &i);
      printf("[");
      print_file_name(&fh);
      printf("]\n");
#if 0
      for(i = 0; i < 5; ++i) {
	printf("%04x ", rec->link.data[i]);
      }
      printf("\n");
#endif
    }
    else {
      printf("record type = 0x%02x\t", rec->link.record_type);
#if 0
      printf("record sub type = 0x%02x\n", rec->normal.record_sub_type);
#endif
      printf("offset = %d ", rec->normal.offset);
      printf("size = %d ", rec->normal.record_size);
      printf("conut = %d\n", rec->normal.continue_count);
#if 0
      printf("address = 0x%06x\n",
	     (rec->normal.logical_block.address.high << 16L) +
	     (rec->normal.logical_block.address.mid << 8L) +
	     rec->normal.logical_block.address.low);
#endif
    }
  }
  else {
    for(i = 0; i < 4; ++i) {
      printf("conut = %d\n", rec->connect.data[i].count);
#if 0
      printf("address = 0x%06x\n",
	     (rec->connect.data[i].logical_block.address.high << 16L) +
	     (rec->connect.data[i].logical_block.address.mid << 8L) +
	     rec->connect.data[i].logical_block.address.low);
#endif
    }
  }
}

void read_file_header(int file_id,
		 FILE_HEADER *file_header, int *file_start_block)
{
  FILE_ID_TABLE id_tbl;

  if ((file_id >= 0) && (file_id <= system_header.max_file)) {
    lseek(fd, 2 * system_header.logical_block_size + 
	  file_id * sizeof(FILE_ID_TABLE), SEEK_SET);
    read(fd, &id_tbl, sizeof(FILE_ID_TABLE));
    *file_start_block = (id_tbl.start_block.address.high << 16L) +
      (id_tbl.start_block.address.mid << 8L) +
      id_tbl.start_block.address.low;
    *file_start_block *= system_header.logical_block_size;
    lseek(fd, *file_start_block, SEEK_SET);
    read(fd, file_header, sizeof(FILE_HEADER));
  }
  else {
    *file_start_block = 0;
  }
}

void read_record(int file_start_block, int recnum, RECORD_INDEX *rec)
{
  lseek(fd,
	file_start_block + recnum * sizeof(RECORD_INDEX)+ RECORD_INDEX_OFFSET,
	SEEK_SET);
  read(fd, rec, sizeof(RECORD_INDEX));
}


void dump_tad(char *buf, int offset, int size)
{
  int i, count;

  for(i = offset, count = 1; i < size; ++i, ++count) {
    printf("%02X ", 0xFF & *buf++);
    if ((count % 16) == 0) printf("\n");
  }
  if (((count-1) % 16) != 0) printf("\n");
}

/* command function */
void do_info()
{
  print_system_header(&system_header);
}

void do_quit()
{
  exit(0);
}

void do_help()
{
  int i;

  for(i = 0; command_tbl[i].command[0] != 0; ++i) {
    printf("%10s", command_tbl[i].command);
    if (((i+1) % 4) == 0) printf("\n");
  }
  if ((i % 4) != 0) printf("\n");
}

void do_finfo()
{
  lseek(fd, file_start_block, SEEK_SET);
  read(fd, &file_header, sizeof(FILE_HEADER));
  print_file_header(&file_header);
}

void do_record()
{
  int i;
  RECORD_INDEX rec;

  printf("record number = %d\n", file_header.record);
  for(i = 0; i < file_header.record; ++i) {
    read_record(file_start_block, i, &rec);
    printf("%2d: ", i);
    print_record_index(&rec);
  }
}

void do_cf()
{
  int i;
  char line[40];

  printf("file id? ");
  fgets(line, 40, stdin);
  sscanf(line, "%d", &i);
  read_file_header(i, &file_header, &file_start_block);
  if (file_start_block != 0) {
    file_id = i;
    printf("file id = %d, file start block = %d\n",
	   file_id, file_start_block);
  }
}

void do_dump()
{
  int num, size, block, count, offset, vsize;
  RECORD_INDEX rec;
  unsigned long address;
  char *buf;

  if (file_header.index_level > 0) {
    printf("Index level higher thant 0 is not supported in this version.\n");
    return;
  }

  for(num = 0; num < file_header.record; ++num) {
    read_record(file_start_block, num, &rec);
    if (rec.normal.record_type == 0x81) break;
  }
  if (rec.normal.record_type != 0x81) {
    printf("There is not main TAD record\n");
    return;
  }

  address = (rec.normal.logical_block.address.high << 16L) +
    (rec.normal.logical_block.address.mid << 8L) +
    rec.normal.logical_block.address.low;
  printf("main tad record = %d count = %d offset = %d block = %x\n", num,
	 rec.normal.continue_count, rec.normal.offset, address);

  buf = alloca(system_header.logical_block_size);
  block = file_header.logical_block;
#if 0
  size = file_header.byte;
#else
  /* Ѥ TAD Ƭ 512 byte ˸ */
  size = 512;
#endif
  count = rec.normal.continue_count;
  offset = rec.normal.offset;
  while(block > 0) {
    lseek(fd, address*system_header.logical_block_size, SEEK_SET);
    read(fd, buf, system_header.logical_block_size);

    vsize = system_header.logical_block_size - offset;
    if (vsize > size) vsize = size;
    dump_tad(buf, offset, vsize);

    /* Ϣ³֥åϥ֥åƬǡޤޤ */
    offset = 0;

    ++address;
    --block;
    --count;
    size -= vsize;
    /* block >0 && count == 0 ʤ顤³ǥå */
    /* ߤ̤б */
    if (count == 0) break;
    if (size <= 0) break;
  }
}

void do_view()
{
  int i, fn, skip;
  int tad_num, lnk_count;
  int count, size, block, offset, vsize;
  RECORD_INDEX rec;
  unsigned long address;
  char *buf;
  TC *tbuf;
  FILE_HEADER *flist;
  FILE_HEADER fh;
  TSS_GEN *tssp;

  if (file_header.index_level > 0) {
    printf("Index level higher thant 0 is not supported in this version.\n");
    return;
  }

  flist = (FILE_HEADER *) alloca(sizeof(FILE_HEADER)*file_header.link_number);
  for(i = 0, lnk_count = 0, tad_num = -1; i < file_header.record; ++i) {
    read_record(file_start_block, i, &rec);
    switch(rec.normal.record_type) {
    case RT_TADDATA:
      /* ǽ˽Ф쥳ɤɽ */
      if (tad_num == -1) tad_num = i;
      break;
    case RT_LINK:
      read_file_header(rec.link.file_id, &fh, &fn);
      memcpy(&(flist[lnk_count]), &fh, sizeof(FILE_HEADER));
      ++lnk_count;
      if (lnk_count > file_header.link_number) {
	printf("illegal link record number\n");
	return;
      }
      break;
    }
  }
  if (tad_num < 0) {
    printf("There is not main TAD record\n");
    return;
  }

  printf("Main TAD record is %d\n", tad_num);
  read_record(file_start_block, tad_num, &rec);
  address = (rec.normal.logical_block.address.high << 16L) +
    (rec.normal.logical_block.address.mid << 8L) +
    rec.normal.logical_block.address.low;
  block = file_header.logical_block;
  size = rec.normal.record_size;
  count = rec.normal.continue_count;
  offset = rec.normal.offset;
  buf = (char *) alloca(size);
  tbuf = (TC *) buf;

  while(block > 0) {
    vsize = system_header.logical_block_size - offset;
    if (vsize > size) vsize = size;

    lseek(fd, address*system_header.logical_block_size + offset, SEEK_SET);
    read(fd, buf, vsize);

    /* Ϣ³֥åϥ֥åƬǡޤޤ */
    offset = 0;
    buf += vsize;
    ++address;
    --block;
    --count;
    size -= vsize;
    if (size <= 0) break;
    /* block >0 && count == 0 ʤ顤³ǥå */
    /* ߤ̤б */
    if (count == 0) {
      printf("Continuous index is not supported in this time\n");
      break;
    }
  }
  
  size = rec.normal.record_size - size;  /* ɤ߻Ĥ */
  for (i = 0; i < size;) {
    switch(*tbuf & 0xFF00) {
    case TC_SPEC:
      tssp = (TSS_GEN *) tbuf;
      switch(tssp->id & 0x00FF) {
      case TST_VOBJ:
	printf(" \033[7m");
	print_file_name(flist++);
	printf("\033[m ");
	break;
      default:
	break;
      }
      skip = tssp->len + sizeof(TSS_GEN);
      break;
    case TC_LANG:
      skip = 2;
      break;
    case TC_NULL:
      switch(*tbuf) {
      case TC_TAB:
	printf("\t"); break;
      case TC_NL:
	printf("\n"); break;
      case TC_FF:
	printf("\n"); break;
      case TC_CR:
	printf("\n"); break;
      case TC_SP:
	printf(" "); break;
      }
      skip = 2;
      break;
    default:
      printk(*tbuf);
      skip = 2;
      break;
    }
    i += skip; tbuf += (skip/2);
  }
  printf("\n");
}
