/*
   statbfs.c
   Copyright (C) 2001 Tomohide Naniwa. All rights reserved.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Last modified: Apr. 17, 2001.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include "btronfs.h"

/* ѿ */
SYSTEM_HEADER system_header;
FILE_HEADER file_header;
int file_id = 0;
int file_start_block = 0;
int fd;
time_t BTRON_TIME;

/* ؿ */
void print_system_header(SYSTEM_HEADER *system_header);
void do_info();
void do_quit();
void do_help();
void do_finfo();
void do_record();
void do_cf();

struct {
  char command[20];
  void (*func)();
} command_tbl[] = {
  {"dinfo", do_info},
  {"finfo", do_finfo},
  {"record", do_record},
  {"cf", do_cf}, 
  {"help", do_help},
  {"quit", do_quit},
  {{0}, NULL}
};

main(int argc, char *argv[])
{
  int i;
  char command[40];
  struct tm btm;

  if (argc < 2) {
    printf("usage: %s 'file name'\n", argv[0]);
    exit(0);
  }

  if ((fd = open(argv[1], O_RDWR)) == -1) {
    perror(argv[1]);
    exit(0);
  }

  /* ƥإåɤ߹ */
  lseek(fd, BLOCK_SIZE, SEEK_SET);
  read(fd, &system_header, sizeof(SYSTEM_HEADER));

  if ((system_header.disk_id != 0x42FE) && (system_header.disk_id != 0x52FE)) {
    printf("File '%s' is not BTRON floppy disk\n", argv[1]);
    exit(1);
  }

  btm.tm_sec = 0;
  btm.tm_min = 0;
  btm.tm_hour = 0;
  btm.tm_mday = 1;
  btm.tm_mon = 0;
  btm.tm_year = 85;
  BTRON_TIME = mktime(&btm);
#ifdef notdef
  printf("BTRON_TIME = %d\n", BTRON_TIME);
  BTRON_TIME = 473353200;
#endif

  /* root file ɤ߹ */
  file_start_block
    = 1024 +
    (system_header.block_size + system_header.id_tbl_size
     + system_header.short_name_tbl_size) *
    system_header.logical_block_size;
  lseek(fd, file_start_block, SEEK_SET);
  read(fd, &file_header, sizeof(FILE_HEADER));

  while(1) {
    printf("statbfs> ");
    fgets(command, 40, stdin);
    command[strlen(command)-1] = 0;
    for(i = 0; command_tbl[i].command[0] != 0; ++i) {
      if (strcmp(command, command_tbl[i].command) == 0) {
	(command_tbl[i].func)();
	break;
      }
    }
  }
}

void printk(unsigned short int wchar)
{
  wchar |= 0x8080;
  printf("%c%c", wchar >> 8, wchar & 0x0ff);
}

void print_system_header(SYSTEM_HEADER *header)
{
  int i;
  time_t clock;

  printf("disk id = 0x%x\n", header->disk_id);
  printf("disk type = 0x%x\n", header->disk_type);
  printf("block size = %d\n", header->block_size);
  printf("id tbl size = %d\n", header->id_tbl_size);
  printf("short name tbl size = %d\n", header->short_name_tbl_size);
  printf("bitmap size = %d\n", header->bitmap_size);
  printf("logical block size = %d\n", header->logical_block_size);
  printf("max file = %d\n", header->max_file);
  printf("lang = 0x%x\n", header->lang);
  printf("access level = %d\n", header->access_level);
  printf("logical block = %d\n", header->logical_block);
  printf("empty bolck = %d\n", header->empty_block);
  clock = header->modified_date + BTRON_TIME;
  printf("modified %s", asctime(gmtime(&clock)));
  clock = header->created_date + BTRON_TIME;
  printf("created %s", asctime(gmtime(&clock)));
  if (header->lang == 0x0021) {
    for(i = 0; i < 20; ++i) {
      if (header->file_system.wchar[i] != 0) {
	printk(header->file_system.wchar[i]);
      }
      else {
	printf("\n");
	break;
      }
    }
    for(i = 0; i < 20; ++i) {
      if (header->device_name.wchar[i] != 0) {
	printk(header->device_name.wchar[i]);
      }
      else {
	printf("\n");
	break;
      }
    }
  }
}

void print_file_header(FILE_HEADER *header)
{
  int i;

  printf("file header id = 0x%x\n", header->header_start_id);
  printf("file type = 0x%04x\n", header->file_type);
  printf("application type = 0x%x\n", header->application_type);
  printf("link number = %d\n", header->link_number);
  printf("index level = %d\n", header->index_level);
  printf("byte = %d\n", header->byte);
  printf("logical block = %d\n", header->logical_block);
  printf("record = %d\n", header->record);
  if (system_header.lang == 0x0021) {
    for(i = 0; i < 20; ++i) {
      if (header->file_name.wchar[i] != 0) {
	printk(header->file_name.wchar[i]);
      }
      else {
	printf("\n");
	break;
      }
    }
  }
  printf("file id = %d\n", header->file_id);
  printf("file end id = 0x%x\n", header->file_end_id);
}

print_record_index(RECORD_INDEX *rec)
{
  int i;

  if (rec->count == 0) {
    if (rec->normal.record_type == 0) {
      printf("\n");
      return;
    }
    else if (rec->normal.record_type == 0x80) {
      printf("record type = 0x%02x\t", rec->link.record_type);
      printf("record sub type = 0x%04x\t", rec->link.record_sub_type);
      printf("file id = %d\n", rec->link.file_id);
#if 0
      for(i = 0; i < 5; ++i) {
	printf("%04x ", rec->link.data[i]);
      }
      printf("\n");
#endif
    }
    else {
      printf("record type = 0x%02x\t", rec->link.record_type);
      printf("record sub type = 0x%02x\n", rec->normal.record_sub_type);
      printf("offset = %d\n", rec->normal.offset);
      printf("size = %d\n", rec->normal.record_size);
      printf("conut = %d\n", rec->normal.continue_count);
      printf("address = 0x%06x\n",
	     (rec->normal.logical_block.address.high << 16L) +
	     (rec->normal.logical_block.address.mid << 8L) +
	     rec->normal.logical_block.address.low);
    }
  }
  else {
    for(i = 0; i < 4; ++i) {
      printf("conut = %d\n", rec->connect.data[i].count);
      printf("address = 0x%06x\n",
	     (rec->connect.data[i].logical_block.address.high << 16L) +
	     (rec->connect.data[i].logical_block.address.mid << 8L) +
	     rec->connect.data[i].logical_block.address.low);
    }
  }
}

/* command function */
void do_info()
{
  print_system_header(&system_header);
}

void do_quit()
{
  exit(0);
}

void do_help()
{
  int i;

  for(i = 0; command_tbl[i].command[0] != 0; ++i) {
    printf("%10s", command_tbl[i].command);
    if (((i+1) % 4) == 0) printf("\n");
  }
  if (((i+1) % 4) != 0) printf("\n");
}

void do_finfo()
{
  lseek(fd, file_start_block, SEEK_SET);
  read(fd, &file_header, sizeof(FILE_HEADER));
  print_file_header(&file_header);
}

void do_record()
{
  int i;
  RECORD_INDEX rec;
  
  lseek(fd, file_start_block+384, SEEK_SET);
  for(i = 0; i < file_header.record; ) {
    read(fd, &rec, sizeof(RECORD_INDEX));
    printf("%2d: ", i);
    print_record_index(&rec);
    if ((rec.count == 0) && (rec.normal.record_type != 0)) ++i;
  }
}

void do_cf()
{
  int i;
  char line[40];
  FILE_ID_TABLE id_tbl;

  printf("file id? ");
  fgets(line, 40, stdin);
  sscanf(line, "%d", &i);
  if ((i >= 0) && (i <= system_header.max_file)) {
    file_id = i;
    lseek(fd, 2 * system_header.logical_block_size + 
	  file_id * sizeof(FILE_ID_TABLE), SEEK_SET);
    read(fd, &id_tbl, sizeof(FILE_ID_TABLE));
    file_start_block = (id_tbl.start_block.address.high << 16L) +
      (id_tbl.start_block.address.mid << 8L) +
      id_tbl.start_block.address.low;
    file_start_block *= system_header.logical_block_size;
    printf("file id = %d, file start block = %d\n", file_id, file_start_block);
    lseek(fd, file_start_block, SEEK_SET);
    read(fd, &file_header, sizeof(FILE_HEADER));
  }
}
